import sys
from hashlib import sha256

# Increase recursion limit
sys.setrecursionlimit(100000)

# Define zero, successcor function, and equality
# These are the only 3 things that will use internal knowledge of the
# model we use

ZERO = b"THIS IS ZERO"


def S(n):
    return sha256(n).digest()


def num_equals(a, b) -> bool:
    return a == b


# Define more numbers
ONE = S(ZERO)
TWO = S(ONE)
THREE = S(TWO)
FOUR = S(THREE)
FIVE = S(FOUR)
SIX = S(FIVE)
SEVEN = S(SIX)
EIGHT = S(SEVEN)
NINE = S(EIGHT)
TEN = S(NINE)

# Add some assertions to check if everything works as it should
assert num_equals(ZERO, ZERO)
assert num_equals(TEN, TEN)
assert not num_equals(TEN, FIVE)
assert not num_equals(FIVE, TEN)
assert num_equals(THREE, S(S(ONE)))
assert num_equals(S(S(SEVEN)), NINE)

# We can now start to define operations on natural numbers
# We will start with comparison functions


# The basic idea for the less operator
# If the numbers are equal, return False
# If the numbers are not, start counting up from 0
# If a is reached at some point, but b is not, a is less than b
# If b is reached at some point, but a is not, a is not less than b
def num_less(a, b) -> bool:
    """Is a less than b (a < b)"""
    if num_equals(a, b):
        return False
    u = ZERO
    while True:
        if num_equals(u, a):
            return True
        if num_equals(u, b):
            return False
        u = S(u)


assert not num_less(FIVE, FIVE)
assert not num_less(TEN, TEN)
assert num_less(FIVE, TEN)
assert not num_less(TEN, FIVE)
assert num_less(ZERO, ONE)


# A number is greater than another number if it's not less and not equal to it
def num_greater(a, b) -> bool:
    """Is a greater than b (a > b)"""
    return (not num_less(a, b)) and (not num_equals(a, b))


assert not num_greater(FIVE, FIVE)
assert not num_greater(TEN, TEN)
assert not num_greater(FIVE, TEN)
assert num_greater(TEN, FIVE)
assert not num_greater(ZERO, ONE)


# Now we can start adding arithmetic operations
# The basic idea for addition
# Set number r to be equal to number a, set number u to 0
# Count up both r and u
# Stop when u reaches b, return r as result
def num_add(a, b):
    """Add numbers a and b (a + b)"""
    r = a
    u = ZERO
    while not num_equals(u, b):
        r = S(r)
        u = S(u)
    return r


assert num_equals(num_add(ZERO, ZERO), ZERO)
assert not num_equals(num_add(ZERO, ONE), ZERO)
assert num_equals(num_add(ZERO, ONE), ONE)
assert not num_equals(num_add(TWO, TWO), FIVE)
assert num_equals(num_add(TWO, TWO), FOUR)
assert num_equals(num_add(THREE, FIVE), EIGHT)
assert num_equals(num_add(FIVE, THREE), EIGHT)


# The basic idea for multiplication
# Set number r to be equal to number a, set number u to 0
# Count up u, on each count, add a to r
# Stop when u reaches b, return r as result
def num_multiply(a, b):
    """Multiply numbers a and b (a * b)"""
    r = ZERO
    u = ZERO
    while not num_equals(u, b):
        r = num_add(r, a)
        u = S(u)
    return r


assert num_equals(num_multiply(ZERO, ZERO), ZERO)
assert num_equals(num_multiply(ZERO, ONE), ZERO)
assert not num_equals(num_multiply(ZERO, ONE), ONE)
assert not num_equals(num_multiply(TWO, TWO), FIVE)
assert num_equals(num_multiply(TWO, TWO), FOUR)
assert num_equals(num_multiply(THREE, TWO), SIX)
assert num_equals(num_multiply(TWO, THREE), SIX)


# The basic idea for subtraction
# If a = b, a - b = 0
# Otherwise, let x = a - b
# Then we need to find such x that x + b = a
# We can do that via repeated additions
def num_subtract(a, b):
    """Subtract number b from a (a - b)"""
    if num_less(a, b):
        raise ValueError("We should stay within natural numbers")
    if num_equals(a, b):
        return ZERO
    x = ZERO
    while not num_equals(num_add(x, b), a):
        x = S(x)
    return x


assert num_equals(num_subtract(ZERO, ZERO), ZERO)
assert num_equals(num_subtract(TEN, TEN), ZERO)
assert num_equals(num_subtract(FIVE, THREE), TWO)
assert not num_equals(num_subtract(FIVE, THREE), THREE)
assert num_equals(num_subtract(FIVE, TWO), THREE)
assert num_equals(num_subtract(TEN, ONE), NINE)
assert not num_equals(num_subtract(TEN, FIVE), SEVEN)


# The basic idea for quotient/modulo
# Let a // b = u
# Then we can find such u that b * S(u) >= a
# Find smallest such u
# If b * S(u) = a, then quotient is S(u), and remainder is 0
# If b * S(u) > a, then quotient is u, and remainder is a - b * u
def num_divmod(a, b):
    """Return the quotient and the remainder a divided by b (a // b, a % b)"""
    if num_equals(b, ZERO):
        raise ValueError("Division by zero")
    u = ZERO
    while num_less(num_multiply(b, S(u)), a):
        u = S(u)
    if num_equals(num_multiply(b, S(u)), a):
        return S(u), ZERO
    return u, num_subtract(a, num_multiply(b, u))


assert num_equals(num_divmod(ZERO, ONE)[0], ZERO)
assert num_equals(num_divmod(ZERO, ONE)[1], ZERO)
assert num_equals(num_divmod(FIVE, TEN)[0], ZERO)
assert num_equals(num_divmod(FIVE, TEN)[1], FIVE)
assert num_equals(num_divmod(TEN, FOUR)[0], TWO)
assert num_equals(num_divmod(TEN, FOUR)[1], TWO)
assert num_equals(num_divmod(FIVE, TWO)[0], TWO)
assert not num_equals(num_divmod(FIVE, TWO)[1], ZERO)
assert num_equals(num_divmod(FIVE, TWO)[1], ONE)


# We can also define a small convenience function to test for divisibility
# Just compute quotient and remainder, then test if remainder is 0
def is_divisible(a, b) -> bool:
    _, r = num_divmod(a, b)
    return num_equals(r, ZERO)


assert is_divisible(TEN, FIVE)
assert not is_divisible(FIVE, TEN)
assert is_divisible(FOUR, TWO)
assert not is_divisible(FIVE, TWO)
assert not is_divisible(TEN, THREE)
assert is_divisible(NINE, THREE)

# Now that we can do enough arithmetic, we can define functions to
# convert between strings and our numbers

# Define a map between decimal digits and numbers within our system
STRING_TO_NUM_MAP = {
    "0": ZERO,
    "1": ONE,
    "2": TWO,
    "3": THREE,
    "4": FOUR,
    "5": FIVE,
    "6": SIX,
    "7": SEVEN,
    "8": EIGHT,
    "9": NINE,
}

# And a reverse map
NUM_TO_STRING_MAP = dict()
for k, v in STRING_TO_NUM_MAP.items():
    NUM_TO_STRING_MAP[v] = k


# The basic idea for conversion from strings to numbers
# Separate the string into individual digits and reverse
# Set u = 0, p = 1
# For each digit d in string set u = u + d * p, p = p * 10
def string_to_num(string: str):
    nums = list(string)
    nums.reverse()
    u = ZERO
    p = ONE
    for n in nums:
        d = STRING_TO_NUM_MAP[n]
        d = num_multiply(d, p)
        u = num_add(u, d)
        p = num_multiply(p, TEN)
    return u


assert num_equals(string_to_num("10"), TEN)
assert num_equals(num_add(string_to_num("10"), FOUR), num_add(SEVEN, SEVEN))
assert num_equals(string_to_num("14"), num_add(SEVEN, SEVEN))
assert num_equals(string_to_num("105"), num_add(num_multiply(TEN, TEN), FIVE))


# The basic idea for converstion from numbers to strings
# Use quotient and remainder repeatetdly to separate the number into
# individual digits
def num_to_string(num) -> str:
    out_string = ""
    while num_greater(num, ZERO):
        num, d = num_divmod(num, TEN)
        out_string = NUM_TO_STRING_MAP[d] + out_string
    return out_string


assert num_to_string(string_to_num("1234")) == "1234"
assert num_to_string(num_add(num_multiply(TEN, TEN), FIVE)) == "105"
assert num_to_string(num_add(SEVEN, SEVEN)) == "14"
assert num_to_string(num_multiply(NINE, THREE)) == "27"


# We can also add a small convenience function (generator) for
# Iterating over consecutive numbers
def num_range(a, b):
    if num_less(b, a):
        raise ValueError("b must not be less than a")
    u = a
    while u != b:
        yield u
        u = S(u)


assert set(num_range(ONE, ONE)) == set()
assert set(num_range(THREE, FIVE)) == set([THREE, FOUR])
assert set(num_range(SEVEN, TEN)) == set([SEVEN, EIGHT, NINE])

if __name__ == "__main__":
    for n in num_range(ONE, string_to_num("101")):
        div3 = is_divisible(n, THREE)
        div5 = is_divisible(n, FIVE)
        if div5 and div3:
            print("FizzBuzz")
        elif div3:
            print("Fizz")
        elif div5:
            print("Buzz")
        else:
            print(num_to_string(n))
